import os 
import typing as ty

from flask import Flask, request, jsonify, g
from flask_login import LoginManager
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from chives.models import Base, Order, Transaction
from chives.configs import environment_overwrite, DEFAULT_CONFIG


login_manager = LoginManager()

def create_app(config_overwrite: ty.Optional[ty.Dict] = None) -> Flask:
    """Create and configure the application

    :param config_overwrite: [description], defaults to None
    :type config_overwrite: [type], optional
    :return: a Flask application
    :rtype: Flask
    """
    # rc is short for runtime configuration
    rc = environment_overwrite(DEFAULT_CONFIG)
    if config_overwrite:
        rc.update(config_overwrite)

    # Create database schema if it doesn't exist
    sql_engine = create_engine(rc['SQLALCHEMY_CONN'])
    Base.metadata.create_all(sql_engine, checkfirst=True)

    # Create the application and set the runtime configuration
    app = Flask(__name__, instance_relative_config=True)
    app.config.from_mapping(rc)

    try:
        os.makedirs(app.instance_path)
    except OSError:
        pass

    from chives.db import init_app 
    init_app(app)
    login_manager.init_app(app)

    from chives.blueprints.auth import bp as auth_bp 
    from chives.blueprints.debug import bp as debug_bp
    from chives.blueprints.exchange import bp as ex_bp
    from chives.blueprints.api import bp as api_bp
    app.register_blueprint(auth_bp)
    app.register_blueprint(debug_bp)
    app.register_blueprint(ex_bp)
    app.register_blueprint(api_bp)
    
    return app



from flask import Blueprint

from flask_login import login_required, current_user

bp = Blueprint("debug", __name__, url_prefix="/debug")

@bp.route("/is_authenticated", methods=("GET",))
@login_required
def is_authenticated():
    return f"You are authenticated as {current_user.username}"



from flask_wtf import FlaskForm as Form
from wtforms import (
    StringField, PasswordField, BooleanField, SubmitField, IntegerField, 
    SelectField, DecimalField, validators)
from wtforms.validators import (InputRequired, Length, EqualTo, NumberRange, 
    Optional, Regexp)

from chives.db import get_db
from chives.models import Company

class RegistrationForm(Form):
    username = StringField(
        label='username', validators=[InputRequired(), Length(min=4, max=25)])
    password = PasswordField(label='password', validators=[
        InputRequired(),
        EqualTo('confirm', message="Passwords must match")]
    )
    confirm = PasswordField("confirm password")
    submit = SubmitField("Sign up")

class LoginForm(Form):
    username = StringField(label="username", validators=[InputRequired()])
    password = PasswordField(label="password", validators=[InputRequired()])
    
    submit = SubmitField("Log in")

class OrderSubmitForm(Form):
    side = SelectField(label="side", choices=[('ask', 'selling'), ('bid', 'buying')], validators=[InputRequired()])
    size = IntegerField(label="size", validators=[InputRequired(), NumberRange(min=1, message="Order size must be positive")])
    security_symbol = StringField(label="security_symbol", validators=[InputRequired()])
    price = DecimalField(label="price", places=2, validators=[NumberRange(min=0.01, message="Specified target price must be positive"), Optional()])
    all_or_none = BooleanField(label="all_or_none")
    immediate_or_cancel = BooleanField(label="immediate_or_cancel")

    submit = SubmitField("Place order")


class StartCompanyForm(Form):
    """There is not need to validate the share price because the company 
    founder will specify a price when he sells

    :param Form: [description]
    :type Form: [type]
    """
    input_cash = DecimalField(label="price", places=2, validators=[NumberRange(min=0.01, message="Initial investment must be positive"), InputRequired()])
    company_symbol = StringField(label="security_symbol", validators=[InputRequired(), Regexp("^[0-9A-Z]{1,10}$", message="Company symbol not valid")])
    company_name = StringField(label="company_name", validators=[InputRequired(), Length(min=1, max=50)])
    size = IntegerField(label="size", validators=[InputRequired(), NumberRange(min=1, message="Share size must be positive")])

    submit = SubmitField("Start company")

    def validate(self):
        supervalidate = Form.validate(self)
        db = get_db()
        if db.query(Company).get(self.company_symbol.data) is not None:
            self.company_symbol.errors.append("Company symbol already taken")
            return False 
        
        return supervalidate



import functools 
import logging

from flask import (
    Blueprint, flash, g as flask_g, redirect, render_template, request, 
    session as flask_session, url_for
)
from flask_login import login_user, logout_user, current_user
from werkzeug.security import check_password_hash, generate_password_hash 

from chives.webserver import login_manager
from chives.db import get_db
from chives.forms import RegistrationForm, LoginForm
from chives.models import User, Asset


# I am not adding file handler because at deployment, I will use an orchestrator 
# to log console output to a file
logger = logging.getLogger("chives.webserver")
chandle = logging.StreamHandler()
formatter = logging.Formatter(
    '[%(asctime)s] {%(filename)s:%(lineno)d} %(levelname)s - %(message)s')
chandle.setFormatter(formatter)
logger.addHandler(chandle)

bp = Blueprint("auth", __name__, url_prefix="/auth")

@login_manager.user_loader 
def load_user(user_id):
    db_session = get_db()
    if user_id is not None:
        loaded_user = db_session.query(User).get(user_id)
        logger.info(f"Loading {loaded_user}")
        return loaded_user
    return None


@login_manager.unauthorized_handler
def unauthorized():
    return redirect(url_for("auth.login"))


@bp.route("/register", methods=("GET", "POST"))
def register():
    form = RegistrationForm(request.form)
    if request.method == "POST" and form.validate_on_submit():
        db_session = get_db()
        if db_session.query(User).filter(
            User.username==form.username.data).first() is not None:
            form.username.errors.append("Username already taken!")
        else:
            new_user: User = User(
                username=form.username.data, 
                password_hash=generate_password_hash(form.password.data))
            db_session.add(new_user)
            db_session.commit()
            # Each user receives $10,000.00 of cash upon registration
            initial_cash_amount = 10000
            initial_cash = Asset(
                owner_id=new_user.user_id,
                asset_symbol="_CASH",
                asset_amount=initial_cash_amount
            )
            db_session.add(initial_cash)
            db_session.commit()
            logger.info(f"{new_user} received ${initial_cash_amount:.2f}")

            return redirect(url_for("auth.login"))

    return render_template("auth/register.html", form=form)


@bp.route("/login", methods=("GET", "POST"))
def login():
    form = LoginForm(request.form)
    if request.method == "POST" and form.validate_on_submit():
        db_session = get_db()
        user = db_session.query(
            User).filter(User.username==form.username.data).first()
        if (user is not None) and check_password_hash(user.password_hash, 
                                                      form.password.data):
            login_user(user)
            return redirect(url_for('exchange.dashboard'))
        else:
            form.password.errors.append(
                "Login failed; check your username and password")
    
    return render_template("auth/login.html", form=form)

@bp.route("/logout")
def logout():
    logout_user()
    return redirect(url_for("auth.login"))



import argparse 
from chives.db import SQLALCHEMY_URI, DEFAULT_SQLALCHEMY_URI

parser = argparse.ArgumentParser(prog="chives")
subparsers = parser.add_subparsers(help="subcommand help", required=True,
# TODO: https://stackoverflow.com/questions/18282403/argparse-with-required-subcommands/18283730
                                    dest='subcommand')
# Arguments that are shared
parser.add_argument("-v", "--verbose",
    help="If used, will provide more detailed output",
    action="store_true")

# Create the parser for the start_engine command
parser_start_engine = subparsers.add_parser('start_engine', 
    help="Start an instance of a matching engine")
parser_start_engine.add_argument("-q", "--queue-host", 
    help="hostname of the RabbitMQ server; defaults to localhost",
    dest="queue_host",
    default="localhost")
parser_start_engine.add_argument("-s", "--sql-uri",
    help=f"Database URI; defaults to {DEFAULT_SQLALCHEMY_URI}",
    dest="sql_uri",
    default=f"{DEFAULT_SQLALCHEMY_URI}")
parser_start_engine.add_argument("--dry-run",
    help="Do not heartbeat the matching engine when a message is received",
    dest="dry_run",
    action="store_true",
    default=False)

# Create the parser for initdb command
parser_initdb = subparsers.add_parser('initdb', 
    help="Initialize the database")
parser_initdb.add_argument("-s", "--sql-uri",
    help=f"Database URI; defaults to {DEFAULT_SQLALCHEMY_URI}",
    dest="sql_uri",
    default=f"{DEFAULT_SQLALCHEMY_URI}")

# Create the parser for webserver command
parser_webserver = subparsers.add_parser('webserver', 
    help="Initialize the database")
parser_webserver.add_argument("-p", "--port",
    help="Flask application port, defaults to 5000",
    dest="webserver_port",
    default="5000")
parser_webserver.add_argument("-d", "--debug",
    help="If used, will make the Flask app run in debug mode",
    dest="debug",
    action="store_true")
parser_webserver.add_argument("-s", "--sql-uri",
    help=f"Database URI; defaults to {DEFAULT_SQLALCHEMY_URI}",
    dest="sql_uri", 
    default=DEFAULT_SQLALCHEMY_URI)



DEFAULT_CONFIG = {
    "SQLALCHEMY_CONN": "sqlite:////tmp/chives.sqlite",
    "SQLALCHEMY_ECHO": False,
    "RABBITMQ_HOST": "localhost",
    "RABBITMQ_PORT": 5672,
    "RABBITMQ_VHOST": "/",
    "RABBITMQ_LOGIN": "guest",
    "RABBITMQ_PASSWORD": "guest",
    "MATCHING_ENGINE_DRY_RUN": False,
    "SECRET_KEY": "dev"
}


import datetime as dt
import logging
import os
import socket
import sys
import time
import typing as ty

import pika
from sqlalchemy import create_engine
from sqlalchemy.engine import Engine as SQLEngine
from sqlalchemy.orm import sessionmaker, Session

from chives.configs import DEFAULT_CONFIG, environment_overwrite
from chives.models import (
    Base, Order, Transaction, Asset, User, Company, MatchingEngineLog)


# I am not adding file handler because at deployment, I will use an orchestrator 
# to log console output to a file
logger = logging.getLogger("chives.matchingengine")
chandle = logging.StreamHandler()
formatter = logging.Formatter(
    '[%(asctime)s] {%(filename)s:%(lineno)d} %(levelname)s - %(message)s')
chandle.setFormatter(formatter)
logger.setLevel(logging.INFO)
logger.addHandler(chandle)


class OrderNotFoundError(KeyError):
    """The exception to raise when the orderbook instance is asked to retrieve 
    an order that does not exist in self.active_orders
    """
    pass


class MatchResult:
        """A dummy class for enforcing a schema for match result
        - incoming is an Order object attached to the main database's session
        - incoming_remain is one of three possibilities:
            - incoming itself
            - a session-less Order object (suborder of incoming)
            - None
        - deactivated is a list of order_id's that 
            - for corresponding entries in main db, change .active to False
            - remove from ob db
        - reactivated is a single sessionless Order object or None
        - transactions is a list of sessionless Transaction objects
        """
        def __init__(self):
            self.incoming: Order = None
            self.incoming_remain: Order = None 
            self.deactivated: ty.List[Order] = []
            self.reactivated: Order = None
            self.transactions: ty.List[Transaction] = []


class MatchingEngine:
    """Abstracting an instance of the matching engine, which wraps around the 
    order book instance and provides an algorithm for matching incoming orders 
    with active orders
    """
    heartbeat_finish_msg = "Heartbeat finished"

    def __init__(self, me_sql_engine: SQLEngine,
                       ignore_user_logic: bool = False,
                       hostname: str = None):
        """Initialize the matching engine by instantiating the order book and 
        creating a engine-bound session

        :param me_sql_engine: The engine for connecting to the main database
        :type me_sql_engine: SQLEngine
        :param ignore_user_logic: if True, the matching engine will not modify 
        user assets after transactions are committed. Defaults to False
        :type ignore_user_logic: bool
        :param hostname: if a hostname is specified, use the specified hostname 
        otherwise, use socket.gethostname(), defaults to None
        :type hostname: str, optional
        """
        Session = sessionmaker(bind=me_sql_engine, autoflush=False)
        self.session = Session()
        self.ignore_user_logic = ignore_user_logic
        self.hostname = hostname if hostname else socket.gethostname()
        self.pid = os.getpid()
    
    def get_order(self, order_id: int) -> Order:
        """Read an order by its order_id

        :param order_id: ID of the order to be obtained
        :type order_id: int
        :return: the order instance
        :rtype: Order
        """
        order = self.session.query(Order).get(order_id)
        if not order:
            raise OrderNotFoundError(f"Order ID {order_id} cannot be found")
        else:
            return order

    def get_candidates(self, incoming: Order) -> ty.List[Order]:
        """Given an incoming order, return all active orders of the same 
        security symbol, that are on the opposite sides, that do not come from 
        the same owner, and that offer better price than the incoming order, 
        if the incoming order has a target price

        :param incoming: the incoming order
        :type incoming: Order
        :return: A list of candidate orders
        :rtype: ty.List[Order]
        """
        candidates: ty.List[Order] = []
        cond = (Order.security_symbol == incoming.security_symbol) \
            & (Order.active == True)
        if incoming.owner_id:
            cond = cond & (Order.owner_id != incoming.owner_id)
        best_price = None
        if incoming.side == "bid":
            cond = cond & (Order.side == "ask")
            if incoming.price:
                cond = cond & (Order.price <= incoming.price)
            best_price = Order.price.asc()
        else:
            cond = cond & (Order.side == "bid")
            if incoming.price:
                cond = cond & (Order.price >= incoming.price)
            best_price = Order.price.desc()
        
        return self.session.query(Order).filter(cond).order_by(
            best_price, Order.create_dttm.desc()).all()

    @classmethod 
    def propose_trade(cls, incoming: Order, 
                           candidate: Order) -> ty.Optional[Transaction]:
        """Given the incoming order and a candidate order, return a Transaction 
        object that describes the trade that can happen between the two orders.
        If no trade can be proposed between the two orders, then None will be
        returned.

        :param incoming: the incoming order
        :type incoming: Order
        :param candidate: the candidate order
        :type candidate: Order
        :return: the proposed trade between the two order, or None if no trade 
        can be proposed
        :rtype: Transaction
        """
        # You can safely assume that incoming and candidate are on opposite 
        # sides and have matching target price, and that the candidate has a 
        # numeric target price
        if incoming.remaining_size <= 0 or candidate.remaining_size <= 0:
            return None 
        else:
            ask: Order = incoming if incoming.side == 'ask' else candidate 
            bid: Order = incoming if incoming.side == 'bid' else candidate
            # with the if statement above, transaction_size is guaranteed to 
            # be non-zero
            transaction_size = min(ask.remaining_size, bid.remaining_size)

            # Check if the candidate is all-or-none and accommodate it
            if candidate.all_or_none \
                and (transaction_size < candidate.remaining_size):
                    # If the candidate is all-or-none and the transaction 
                    # cannot fulfill it entirely
                    logger.debug(
                        f"AON resting order {candidate} rejected {incoming}")
                    return None
            else:
                # the transaction price will always be the candidate's 
                # price since by prior filtering, the candidate will 
                # always have the better pricing

                transaction: Transaction = Transaction(
                    security_symbol=ask.security_symbol,
                    size=transaction_size,
                    price=candidate.price,
                    ask_id=ask.order_id,
                    bid_id=bid.order_id,
                    aggressor_order_id=incoming.order_id,
                    resting_order_id=candidate.order_id
                )

                return transaction
    
    def exchange_user_asset(self, transaction: Transaction):
        """Given a transaction, find the parties involved in the transaction, 
        then perform the exchange of cash for stocks as reflected on the assets
        table

        :param transaction: [description]
        :type transaction: Transaction
        """
        # Identify the seller and the buyer.
        ask = self.session.query(Order).get(transaction.ask_id)
        bid = self.session.query(Order).get(transaction.bid_id)
        seller: User = ask.owner
        buyer: User = bid.owner

        # Seller gains cash and loses stocks, but the stock share 
        # deduction happens at order's submission, so we only need 
        # to be concerned with gaining cash.
        cash_volume = transaction.price * transaction.size 
        seller_cash: Asset = self.session.query(Asset).get(
            (seller.user_id, "_CASH")
        )
        seller_cash.asset_amount += cash_volume
        self.session.merge(seller_cash)

        # Buyer gains stock and loses cash, both of which will 
        # happen here, since there is no telling what price the 
        # buyer will get when the buyer places an order.
        #
        # First check if the buyer has had this kind of asset before
        buyer_stock = self.session.query(Asset).get(
            (buyer.user_id, transaction.security_symbol)
        )
        # If the buyer has this kind of asset, then add to it; 
        # otherwise, create a new entry
        if buyer_stock is None:
            self.session.add(
                Asset(owner_id=buyer.user_id,
                        asset_symbol=transaction.security_symbol,
                        asset_amount=transaction.size))
        else:
            buyer_stock.asset_amount += transaction.size
            self.session.merge(buyer_stock)
        # Remove cash from buyer
        buyer_cash = self.session.query(Asset).get(
            (buyer.user_id, "_CASH")
        )
        buyer_cash.asset_amount -= cash_volume
        self.session.merge(buyer_cash)

    def update_market_price(self, transaction: Transaction):
        """Given a transaction object, find the company that this transaction 
        is trading on, then update the company's market price with the 
        transaction price

        :param transaction: [description]
        :type transaction: Transaction
        """
        # Update the company's market price
        company = self.session.query(Company).get(
            transaction.security_symbol)
        company.market_price = transaction.price
        self.session.merge(company)

    def refund_cancelled_remains(self, match_result: MatchResult):
        """If the incoming order is a selling order that is not entirely 
        fulfilled, and whose remaining part is cancelled, then return the 
        remaining part's assets back to the seller

        :param match_result: [description]
        :type match_result: MatchResult
        """
        if (match_result.incoming.side == "ask") \
            and match_result.incoming_remain is not None \
            and (match_result.incoming_remain.cancelled_dttm is not None):
            owner_id = match_result.incoming_remain.owner_id
            symbol = match_result.incoming_remain.security_symbol
            source_asset = self.session.query(Asset).get((owner_id, symbol))
            refund_size = match_result.incoming_remain.size
            logger.debug(f"Refunding {refund_size} shares")
            source_asset.asset_amount += refund_size
            self.session.merge(source_asset)

    def process_match_result(self, match_result: MatchResult):
        """Write changes described by the match result into the database:
        1.  incoming_order needs to be merged because it might be cancelled 
            or be made active
        2.  incoming_remain, if distinct from incoming_order and not None, 
            needs to be added into the database as a new order
        3.  A number of resting orders need to be deactivated because they 
            matched with the aggressor order and produced transaction
        4.  If there is a sub-order of a resting order, then it needs to be 
            added as a new order
        5.  For each transaction, add it into the database. 
            If user logic is not to be ignored, then 
            *   call exchange_user_asset to execute with the exchange of 
                assets induced by said transaction
            *   call update_market_price to update a company's stock's market
                price
        6.  If user logic is not to be ignored, then refund unfulfilled remains 
            of the selling order back to the user appropriately

        :param match_result: [description]
        :type match_result: MatchResult
        """
        self.session.merge(match_result.incoming)
        
        if match_result.incoming_remain is not match_result.incoming\
            and match_result.incoming_remain is not None:
            self.session.add(match_result.incoming_remain)
        
        if len(match_result.deactivated) > 0:
            for deactivated in match_result.deactivated:
                main_entry = self.session.query(Order).get(deactivated)
                main_entry.active = False 

        if match_result.reactivated is not None:
            self.session.add(match_result.reactivated)

        if len(match_result.transactions) > 0:
            for transaction in match_result.transactions:
                self.session.add(transaction)
                if not self.ignore_user_logic:
                    self.exchange_user_asset(transaction)
                    self.update_market_price(transaction)
        
        if not self.ignore_user_logic:
            self.refund_cancelled_remains(match_result)
    
    def _heartbeat(self, incoming: Order):
        """Register the incoming order into the main database, run it against 
        self.match, then commit the changes to main database and/or the 
        orderbook database

        :param incoming: the incoming order
        :type incoming: Order
        """
        logger.debug("Starting new heartbeat")
        self.session.close(); time.sleep(0.01)

        # The self.match method does not commit any actual changes to any 
        # database. Instead, it returns the set of changes that need to be 
        # committed.
        match_result: MatchResult = self.match(incoming)
        self.process_match_result(match_result)
        
        self.log_to_sql(msg=self.heartbeat_finish_msg)
        # This is the only commit that will happen for each heartbeat
        self.session.commit()

    def heartbeat(self, incoming: Order):
        try:
            logger.info(f"Trying to heartbeat {incoming}")
            self._heartbeat(incoming)
            logger.info(f"Heartbeated {incoming}")
        except Exception as e:
            logger.error(f"Commit failed: {e}")
            self.session.rollback()
            self.heartbeat(incoming)

    def match(self, incoming: Order) -> MatchResult:
        """The specific logic is recorded in the module README.

        :param incoming: [description]
        :type incoming: Order
        :return: [description]
        :rtype: MatchResult
        """
        mr = MatchResult()
        incoming.remaining_size = incoming.size 
        candidates: ty.List[Order] = self.get_candidates(incoming)
        logger.debug(f"Found {len(candidates)} resting orders as candidates")

        for candidate in candidates:
            candidate.remaining_size = candidate.size

            if incoming.remaining_size <= 0:
                break
            else:
                # candidate's AON policy is respected within propose_trade;
                # if candidate is AON and incoming.remaining_size < candidate.size
                # then transaction is None
                transaction = self.propose_trade(incoming, candidate)
                if transaction is not None:
                    logger.debug(f"{incoming} matches {candidate}: {transaction}")
                    incoming.remaining_size -= transaction.size 
                    candidate.remaining_size -= transaction.size 
                    mr.transactions.append(transaction)
                    mr.deactivated.append(candidate.order_id)
                    # If the candidate is partially fulfilled, then create its 
                    # remains as a suborder
                    if candidate.remaining_size > 0:
                        mr.reactivated = candidate.create_suborder()
        
        # After all matchings are complete
        mr.incoming = incoming
        if incoming.remaining_size == incoming.size:
            logger.debug(f"No trade proposed; incoming order's remain is itself")
            mr.incoming_remain = mr.incoming
            mr.incoming_remain.active = mr.incoming.active = True
        elif incoming.remaining_size > 0:
            logger.debug(f"Incoming order partially fulfilled; creating suborder")
            mr.incoming_remain = incoming.create_suborder()
            mr.incoming_remain.active = True
        else:
            logger.debug(f"Incoming order completely fulfilled")
            mr.incoming_remain = None

        # Respect the AON and IOC policy
        if incoming.all_or_none and incoming.remaining_size > 0:
            logger.debug(f"Incoming order is all-or-none and less than completely fulfilled")
            mr.incoming_remain = mr.incoming
            mr.incoming_remain.active = True
            mr.transactions = []
            # There is no need to refresh the candidate since each match() 
            # reads a fresh record from the database, which means the mutations 
            # in one match cycle does not persist to the next
            mr.deactivated = []
            mr.reactivated = None
        if incoming.immediate_or_cancel and (mr.incoming_remain is not None):
            logger.debug(f"Incoming order is immediate-or-cancel and has non-trivial remains")
            mr.incoming_remain.cancelled_dttm = dt.datetime.utcnow()
            mr.incoming_remain.active = False
        
        return mr

    def log_to_sql(self, msg: str, ext_ref: str = None, ext_ref_id: int = None):
        """Add a log message to the me_logs table

        :param msg: the message; if the message is longer than 1024 characters, 
        then it will be truncated
        :type msg: str
        :param ext_ref: table name of an external record, defaults to None
        :type ext_ref: str, optional
        :param ext_ref_id: id of the external record, defaults to None
        :type ext_ref_id: int, optional
        """
        msg = msg[:1024]
        self.session.add(MatchingEngineLog(
            hostname=self.hostname,
            pid=self.pid,
            log_msg=msg,
            ext_ref=ext_ref,
            ext_ref_id=ext_ref_id
        ))


def start_engine(config_overwrite: ty.Optional[ty.Dict] = None):
    """Obtain the final runtime configuration, then use it to spawn the 
    necessary components and start listening for incoming messages

    :param config_overwrite: overwriting runtime configuration
    :type config_overwrite: dict
    """
    # rc is short for runtime configuration
    rc = environment_overwrite(DEFAULT_CONFIG)
    if config_overwrite:
        rc.update(config_overwrite)

    logger.info("Starting matching engine")

    # Connect to RabbitMQ
    mq_creds = pika.PlainCredentials(
        rc['RABBITMQ_LOGIN'], rc['RABBITMQ_PASSWORD'])
    mq_conn_params = pika.ConnectionParameters(
        rc['RABBITMQ_HOST'], rc['RABBITMQ_PORT'], rc['RABBITMQ_VHOST'], mq_creds
    )
    mq_conn = pika.BlockingConnection(mq_conn_params)
    mq_ch = mq_conn.channel()
    mq_ch.queue_declare(queue="incoming_order")
    logger.info("Connected to RabbitMQ")
    
    # Connect to SQL and try to initialize schema if it does not exists already
    sql_engine = create_engine(
        rc['SQLALCHEMY_CONN'], echo=rc['SQLALCHEMY_ECHO'])
    Base.metadata.create_all(sql_engine, checkfirst=True)

    # Create the engine object
    me = MatchingEngine(sql_engine)
    logger.info(f"Matching engine started at {me.hostname} with pid {me.pid}")

    def msg_callback(ch, method, properties, body):
        logger.info("Received %r" % body)
        if not rc['MATCHING_ENGINE_DRY_RUN']:
            me.heartbeat(Order.from_json(body))
        ch.basic_ack(delivery_tag=method.delivery_tag)
    
    # Tells RabbitMQ not to give more than one message at a time;
    # do not dispatch a new message to a worker until it has processed and 
    # acknowledged the previous one
    mq_ch.basic_qos(prefetch_count=1) 
    mq_ch.basic_consume(queue="incoming_order", 
                     on_message_callback=msg_callback)
    logger.info("Listening for incoming order")
    mq_ch.start_consuming()            



import datetime as dt
from collections import namedtuple
import logging
import os 
import random
import time
import typing as ty

import pika
from sqlalchemy import create_engine, func
from sqlalchemy.engine import Engine as SQLEngine
from sqlalchemy.orm import sessionmaker, Session
from werkzeug.security import generate_password_hash

from chives.models import (
    Base, User, Company, Asset, Order, Transaction, MatchingEngineLog)
from chives.matchingengine import MatchingEngine


DEFAULT_SQLITE_URI = "sqlite:////tmp/benchmark.chives.sqlite"
DEFAULT_MYSQL_URI = "mysql+pymysql://chives_u:chives_password@localhost:3307/chives"
BenchmarkResult = namedtuple("BenchmarkResult", ["run_seconds", "errors"])

logger = logging.getLogger("chives.benchmark")
logger.setLevel(logging.INFO)
chandle = logging.StreamHandler()
chandle.setLevel(logging.INFO)
formatter = logging.Formatter(
    '[%(asctime)s] {%(filename)s:%(lineno)d} %(levelname)s - %(message)s')
chandle.setFormatter(formatter)
logger.addHandler(chandle)


def add_user(username: str, password: str, sql_session: Session) -> User:
    """Add a user with the username and password, then return the user object;
    if something goes wrong, let the commit() method raise exception

    :param username: [description]
    :type username: str
    :param password: [description]
    :type password: str
    :param sql_session: [description]
    :type sql_session: Session
    """
    new_user = User(
        username=username, password_hash=generate_password_hash(password))
    sql_session.add(new_user)
    sql_session.commit()

    return new_user


def add_company(symbol: str, founder_id: int, sql_session: Session) -> Company:
    """Add a new company with the specified symbol and founder_id

    :param symbol: [description]
    :type symbol: str
    :param founder_id: [description]
    :type founder_id: int
    :return: [description]
    :rtype: Company
    """
    new_company = Company(
        symbol=symbol,
        name=symbol,
        initial_value=10000,
        initial_size=10000,
        founder_id=founder_id,
        market_price=1
    )
    sql_session.add(new_company)
    sql_session.commit()

    return new_company


def inject_asset(user_id: int, symbol: str, amount: int, session: Session):
    """Given a user_id, insert a record of asset for this user, then return 
    the Asset object

    :param user_id: [description]
    :type user_id: [type]
    """
    existing_asset = session.query(Asset).get((user_id, symbol))
    if existing_asset is None:
        session.add(Asset(owner_id=user_id, 
            asset_symbol=symbol, asset_amount=amount))
    else:
        existing_asset.asset_amount += amount
    session.commit()

    return session.query(Asset).get((user_id, symbol))


def order_tracing(session: Session) -> ty.List[str]:
    """Perform order/transaction integrity verification and return a list of 
    error messages each describing an inconsistency

    :param session: [description]
    :type session: Session
    :return: [description]
    :rtype: ty.List[str]
    """
    errors = []
    session.expire_all()
    n_orders = session.query(Order).count()
    logger.info(f"Inspecting {n_orders} orders")
    for i in range(n_orders):
        inspected_order: Order = session.query(Order).get(i+1)
        if inspected_order.side == "bid":
            trns_filter = Transaction.bid_id == inspected_order.order_id
        else:
            trns_filter = Transaction.ask_id == inspected_order.order_id
        trades = session.query(Transaction).filter(trns_filter).all()
        trade_volume = sum([t.size for t in trades])
        
        if trade_volume == inspected_order.size:
            logger.debug(f"{inspected_order} is fully traded")
        elif trade_volume > inspected_order.size:
            error_msg = f"{inspected_order} is over-traded with volume {trade_volume}"
            logger.warning(error_msg)
            errors.append(error_msg)
        elif trade_volume > 0:
            # inspected_order is partially traded so we look for sub-order 
            suborder = session.query(Order).filter(
                Order.parent_order_id == inspected_order.order_id).first()
            if suborder is None:
                error_msg = f"{inspected_order} is partially fulfilled but no suborder is found"
                logger.warning(error_msg)
                errors.append(error_msg)
            else:
                if suborder.size + trade_volume != inspected_order.size:
                    error_msg = f"{inspected_order}, its remains {suborder}, and trade volume {trade_volume} are inconsistent"
                    logger.warning(error_msg)
                    errors.append(error_msg)
                elif inspected_order.immediate_or_cancel \
                    and suborder.cancelled_dttm is None:
                    error_msg = f"{inspected_order} is IOC, but its remains {suborder} is not cancelled"
                    logger.warning(error_msg)
                    errors.append(error_msg)
            logger.debug(f"{inspected_order} is partially fulfilled and its remaining is correct")
        else:
            # inspected_order is not traded at all 
            if (not inspected_order.active) \
                and (inspected_order.cancelled_dttm is None):
                error_msg = f"{inspected_order} is not traded at all, but it is neither active nor cancelled"
                logger.warning(error_msg)
                errors.append(error_msg)
            else:
                logger.debug(f"{inspected_order} is not traded at all")
    
    return errors


def _benchmark(sql_session: Session, 
        queue_conn: pika.BlockingConnection, n_rounds: int) -> ty.Tuple[
            dt.datetime, ty.List[int], ty.List[float]]:
    """The core logic of the benchmark: add pseudo users and bench company, 
    inject assets, generate random sizes and prices, create stock order objects, 
    then submit them to the queue

    :param sql_session: [description]
    :type sql_session: Session
    :param queue_conn: [description]
    :type queue_conn: pika.BlockingConnection
    :param n_rounds: [description]
    :type n_rounds: int
    :return: A start datetime
    :rtype: dt.datetime
    """
    # Open Message Queue channel
    ch = queue_conn.channel()
    ch.queue_declare(queue='incoming_order')

    # Set up initial data
    buyer = add_user("buyer", "password", sql_session)
    seller = add_user("seller", "password", sql_session)
    bench_company = add_company("BENCH", seller.user_id, sql_session)
    inject_asset(seller.user_id, "_CASH", 10000000, sql_session)
    inject_asset(buyer.user_id, "_CASH", 10000000, sql_session)
    random_sizes = [random.randint(1, 100) for i in range(n_rounds)]
    random_prices = [random.uniform(10, 100) for i in range(n_rounds)]

    start_dttm = dt.datetime.utcnow()
    order_messages = []
    for i in range(n_rounds):
        random_size, random_price = random_sizes[i], random_prices[i]
        injected_asset = inject_asset(
            seller.user_id, bench_company.symbol, random_size, sql_session)
        injected_asset.asset_amount -= random_size
        sql_session.commit()
        ask = Order(
            security_symbol=bench_company.symbol,
            side="ask",
            size=random_size,
            price=random_price,
            owner_id=seller.user_id
        )
        bid = Order(
            security_symbol=bench_company.symbol,
            side="bid",
            size=random_size,
            price=None,
            owner_id=buyer.user_id,
            immediate_or_cancel=True
        )
        sql_session.add(ask)
        sql_session.add(bid)
        sql_session.commit()
        order_messages.append(ask.json)
        order_messages.append(bid.json)
    
    for msg in order_messages:
        ch.basic_publish(
            exchange='', routing_key='incoming_order', body=msg)
    
    return start_dttm, random_sizes, random_prices


def benchmark(n_rounds: int = 1, sql_uri: str = DEFAULT_SQLITE_URI, 
              verify_integrity: bool = True):
    """Remove existing benchmark.chives.sqlite, create a new one, initialize
    database schema, create buyer/seller/company, then for each round, submit 
    an order into the rabbitMQ. After all rounds, wait until transaction 
    n_rounds exists, then report correctness and runtime.

    This method is not responsible (and is unable to) for spawnning matching 
    engines

    :param sqlite_path: [description], defaults to DEFAULT_SQLITE_PATH
    :type sqlite_path: str, optional
    :param n_workers: [description], defaults to 1
    :type n_workers: int, optional
    :param n_rounds: [description], defaults to 1
    :type n_rounds: int, optional
    :param verify_integrity: if set to True, stops the benchmark after all orders are 
    submitted, then return a BenchmarkResult with 0 second runtime and error 
    message "dry run"
    """
    # Set up database schema
    logger.info(f"""Starting benchmark session: 
    number of rounds: {n_rounds}
    Database URI: {sql_uri}""")
    main_engine = create_engine(sql_uri)
    Base.metadata.drop_all(main_engine) # it's okay to repeatedly run drop_all
    Base.metadata.create_all(main_engine)
    Session = sessionmaker(bind=main_engine)
    main_session = Session()
    logger.info(f"""Database schemas dropped and re-created""")

    # Set up RabbitMQ connection
    mq = pika.BlockingConnection(
        pika.ConnectionParameters(host='localhost'))
    logger.info(f"RabbitMQ connected")

    # Start the initiation part of the benchmark
    start_dttm, random_sizes, random_prices = _benchmark(
        main_session, mq, n_rounds)
    logger.info("All order messages submitted to queue")
    # Once all messages have been submitted, the connection to the rabbitMQ 
    # should be closed immediately; according to the documentation:
    # https://pika.readthedocs.io/en/stable/examples/heartbeat_and_blocked_timeouts.html
    # blocking connection times out after 60 seconds, which means that trying 
    # to close a connection after the timeout will result in an error:
    mq.close()
    logger.info("RabbitMQ connection gracefully closed")
    
    if not verify_integrity:
        # Finish the benchmark without computing runtime or correctness
        logger.info("Skipped integrity verification. Benchmark finished")
        return BenchmarkResult(0, ["Skipped verifying integrity"])
    else:
        # Wait until there are 2 * n_rounds "heartbeat_finished" messages
        hbfinished = MatchingEngineLog.log_msg == MatchingEngine.heartbeat_finish_msg
        while main_session.query(MatchingEngineLog)\
            .filter(hbfinished).count() < (2 * n_rounds):
            main_session.close(); time.sleep(1)
        # Get the last of all heartbeat_finish message
        log_dttm_desc = MatchingEngineLog.log_dttm.desc()
        latest_log = main_session.query(MatchingEngineLog)\
            .filter(hbfinished).order_by(log_dttm_desc).first()
        run_seconds = (latest_log.log_dttm - start_dttm).total_seconds()
        
        error_msgs = order_tracing(main_session)
        logger.info(f"Benchmark finished; {len(error_msgs)} inconsistencies found")

        return BenchmarkResult(run_seconds, error_msgs)



import os
import logging

import click 
from flask import current_app, g, _app_ctx_stack
from flask.cli import with_appcontext
import pika
from sqlalchemy import create_engine 
from sqlalchemy.ext.declarative import declarative_base 
from sqlalchemy.orm import sessionmaker, scoped_session

DEFAULT_SQLALCHEMY_URI = "sqlite:////tmp/chives.sqlite"
SQLALCHEMY_URI = os.getenv("SQLALCHEMY_URI", "sqlite:////tmp/chives.sqlite")

# engine = create_engine(SQLALCHEMY_URI)
# Session = sessionmaker(autocommit=False, autoflush=False, bind=engine)
logger = logging.getLogger("chives.webserver")
chandle = logging.StreamHandler()
formatter = logging.Formatter(
    '[%(asctime)s] {%(filename)s:%(lineno)d} %(levelname)s - %(message)s')
chandle.setFormatter(formatter)
logger.addHandler(chandle)
Base = declarative_base()


def get_db():
    if 'db_session' not in g:
        engine = create_engine(
            current_app.config['SQLALCHEMY_CONN'], 
            echo=current_app.config['SQLALCHEMY_ECHO'])
        Session = sessionmaker(autocommit=False, autoflush=False, bind=engine)
        g.db_session = scoped_session(
            Session, scopefunc=_app_ctx_stack.__ident_func__)
        logger.debug(f"Spawned scoped session for webserver")
    return g.db_session

def close_db(e=None):
    db_session = g.pop('db_session', None)
    if db_session is not None:
        db_session.remove()
    logger.debug(f"ORM Session closed")

def get_mq() -> pika.BlockingConnection:
    """If the webserver application has no active connection to the RabbitMQ 
    server, then create one and append it to the flask G. Otherwise, return 
    the connection on the flask G

    :return: an active connection
    :rtype: pika.BlockingConnection
    """
    if 'mq_conn' not in g:
        app_config = current_app.config
        mq_creds = pika.PlainCredentials(
            app_config['RABBITMQ_LOGIN'], app_config['RABBITMQ_PASSWORD'])
        mq_conn_params = pika.ConnectionParameters(
            app_config['RABBITMQ_HOST'], app_config['RABBITMQ_PORT'], 
            app_config['RABBITMQ_VHOST'], mq_creds
        )
        g.mq_conn = pika.BlockingConnection(mq_conn_params)
        logger.debug(f"Spawned RabbitMQ connection")
    return g.mq_conn

def close_mq(e=None):
    """Close the MQ connection; this method will be added to teardown_appcontext

    :param e: [description], defaults to None
    :type e: [type], optional
    """
    mq_conn: pika.BlockingConnection = g.pop('mq_conn', None)
    if mq_conn is not None:
        mq_conn.close()
    logger.debug(f"Closed RabbitMQ connection")


def init_app(app):
    app.teardown_appcontext(close_db)
    app.teardown_appcontext(close_mq)



import logging

from babel.numbers import format_number
from flask import (
    Blueprint, flash, g as flask_g, redirect, render_template, request, 
    session as flask_session, url_for
)
from flask_login import login_required, current_user
import pika
from pika.exceptions import AMQPConnectionError

from chives.db import get_db, get_mq
from chives.forms import OrderSubmitForm, StartCompanyForm
from chives.models import Order, Asset, Company, Transaction, User

logger = logging.getLogger("chives.webserver")
chandle = logging.StreamHandler()
formatter = logging.Formatter(
    '[%(asctime)s] {%(filename)s:%(lineno)d} %(levelname)s - %(message)s')
chandle.setFormatter(formatter)
logger.addHandler(chandle)
bp = Blueprint("exchange", __name__, url_prefix="/exchange")

@bp.route("/dashboard", methods=("GET",))
@login_required
def dashboard():
    db = get_db()
    # Any amount of cash will be displayed
    cash = [a for a in current_user.assets if (a.asset_symbol == "_CASH")][0]
    cash.asset_amount_display = format_number(cash.asset_amount, locale="en_US")
    # Empty assets for stocks will not be displayed
    stocks = [a for a in current_user.assets 
        if (a.asset_symbol != "_CASH") and (a.asset_amount > 0)]

    # Compute stocks' market value by company's market price * asset_amount
    # Then compute net worth by summing up cash and stocks market value
    net_worth = cash.asset_amount 
    for stock in stocks:
        market_price = db.query(Company).get(stock.asset_symbol).market_price
        stock.market_value = market_price * stock.asset_amount
        stock.market_value_display = format_number(stock.market_value, locale="en_US")
        stock.asset_amount_display = format_number(
            int(stock.asset_amount), locale="en_US")
        net_worth += stock.market_value
    net_worth_display = format_number(net_worth, locale="en_US")
    return render_template("exchange/dashboard.html", 
        cash=cash, stocks=stocks, net_worth=net_worth_display, title="Dashboard")

@bp.route("/submit_order", methods=("GET", "POST"))
@login_required
def submit_order():
    # Even before rendering the page, try to connect to RabbitMQ. If RabbitMQ 
    # is not running, then redirect to an error page
    mq = ch = None
    try:
        mq: pika.BlockingConnection = get_mq()
        ch = mq.channel()
        ch.queue_declare(queue='incoming_order')
    except AMQPConnectionError as e:
        error_msg = "Webserver failed to connect to order queue"
        return redirect(url_for("exchange.error", error_msg=error_msg))

    form = OrderSubmitForm(request.form)
    if request.method == "POST" and form.validate_on_submit():
        new_order: Order = Order(
            security_symbol=form.security_symbol.data, 
            side=form.side.data, 
            size=form.size.data, 
            price=form.price.data,
            all_or_none=form.all_or_none.data,
            immediate_or_cancel=form.immediate_or_cancel.data,
            owner_id=current_user.user_id
        )
        if new_order.side == "ask":
            user_existing_asset = get_db().query(Asset).get(
                (current_user.user_id, new_order.security_symbol))
            if user_existing_asset is None or user_existing_asset.asset_amount < new_order.size:
                return redirect(url_for("exchange.error", error_msg="Insufficient assets"))
            else:
                logger.info(f"Subtracting {new_order.size} shares of {new_order.security_symbol} from {current_user}")
                user_existing_asset.asset_amount -= new_order.size
        if new_order.price is None:
            logger.info(f"Marking market order {new_order}")
            new_order.immediate_or_cancel = True
        db = get_db()
        db.add(new_order)
        db.commit()
        logger.info(f"{new_order} committed to database")

        ch.basic_publish(
                exchange='', routing_key='incoming_order', body=new_order.json)
        logger.info(f"{new_order} submitted to order queue")

        return redirect(url_for("exchange.dashboard"))
    return render_template(
        "exchange/submit_order.html", form=form, title="Submit order")


@bp.route("/recent_orders", methods=("GET",))
@login_required 
def recent_orders():
    """Render the most recent (up to) 50 orders
    """
    db = get_db()
    ownership = (Order.owner_id == current_user.user_id)
    create_dttm_desc = Order.create_dttm.desc()
    recent_orders = db.query(Order).filter(
        ownership).order_by(create_dttm_desc).limit(50).all()
    for order in recent_orders:
        order.side_display = "Buy" if order.side == "bid" else "Sell"
        order.price_display = f"${order.price:.2f}" if order.price is not None else "any price available"
        order.create_dttm_display = order.create_dttm.strftime("%Y-%m-%d %H:%M:%S")
    
    return render_template(
        "exchange/recent_orders.html", orders=recent_orders, title="Recent orders")


@bp.route("/recent_transactions", methods=("GET",))
@login_required
def recent_transactions():
    """Render the most recent (up to) 50 transactions
    """
    db = get_db()
    order_ids = [o.order_id for o in current_user.orders]
    involves_current_user = Transaction.ask_id.in_(order_ids) \
        | Transaction.bid_id.in_(order_ids)
    dttm_desc = Transaction.transact_dttm.desc()
    transactions = db.query(Transaction).filter(
        involves_current_user).order_by(dttm_desc).limit(50).all()

    for t in transactions:
        t.side_display = "Bought" if (t.bid_id in order_ids) else "Sold"
        t.dttm_display = t.transact_dttm.strftime("%Y-%m-%d %H:%M:%S")
    
    return render_template(
        "exchange/recent_transactions.html", 
        transactions=transactions, title="Recent transactions")


@bp.route("/view_company/<company_symbol>", methods=("GET",))
@login_required 
def view_company(company_symbol):
    db = get_db()
    company = db.query(Company).get(company_symbol)
    if company is None:
        return redirect(url_for(
            "exchange.error", 
            error_msg=f"Company {company_symbol} does not exist"))
    else:
        company.create_date_display = company.create_dttm.strftime("%Y-%m-%d")
        company.founder_name = company.founder.username
        return render_template("exchange/view_company.html", 
            company=company, title=f"Company: {company_symbol}")


@bp.route("/start_company", methods=("GET", "POST"))
@login_required 
def start_company():
    form = StartCompanyForm(request.form)
    db = get_db()
    if request.method == "POST" and form.validate_on_submit():
        new_company: Company = Company(
            symbol=form.company_symbol.data,
            name=form.company_name.data,
            initial_value=form.input_cash.data,
            initial_size=form.size.data,
            founder_id=current_user.user_id,
            market_price=form.input_cash.data / form.size.data
        )
        db.add(new_company)

        founder_cash = db.query(Asset).get((current_user.user_id, "_CASH"))
        founder_cash.asset_amount -= float(form.input_cash.data)

        # Since this is a new company, the founder will definitely not have 
        # prior assets
        founder_stocks = Asset(
            owner_id=current_user.user_id,
            asset_symbol=form.company_symbol.data,
            asset_amount=form.size.data)
        db.add(founder_stocks)
        db.commit()
        logger.info(f"{new_company} committed to database")
        logger.info(f"{founder_stocks} committed to database")

        return redirect(url_for("exchange.dashboard"))
    return render_template(
        "exchange/start_company.html", form=form, title="Start company")


@bp.route("/error", methods=("GET",))
def error():
    error_msg = "Something went wrong..."
    if 'error_msg' in request.args:
        error_msg = request.args['error_msg']
    return render_template("exchange/error.html", error_msg=error_msg, title="Error")



import datetime as dt
import json

from flask_login import UserMixin
from sqlalchemy import (
    Column, Integer, String, Boolean, Float, DateTime, ForeignKey)
from sqlalchemy.orm import relationship

from chives.db import Base


class User(UserMixin, Base):
    __tablename__ = 'users'

    user_id = Column(Integer, primary_key=True)
    username = Column(String(120), unique=True, nullable=False)
    password_hash = Column(String(120), nullable=False)

    assets = relationship(
        "Asset", back_populates="owner", 
        cascade="all, delete",
        # setting passive_deletes to True means that the deletion cascade is 
        # handled by the database via the "ON DELETE" declaration, instead by 
        # SQLAlchemy
        passive_deletes=True)
    orders = relationship(
        "Order", back_populates="owner", 
        cascade="all, delete",
        passive_deletes=True)
    companies = relationship(
        "Company", back_populates="founder", 
        cascade="all, delete",
        passive_deletes=True)

    def __str__(self):
        return f"<User user_id={self.user_id}, username={self.username}>"

    def __repr__(self):
        return self.__str__()
    
    def get_id(self):
        return self.user_id


class Asset(Base):
    __tablename__ = 'assets'

    owner_id = Column(Integer, ForeignKey('users.user_id', ondelete="CASCADE"), 
                        primary_key=True)
    asset_symbol = Column(String(10), primary_key=True)
    asset_amount = Column(Float, nullable=False)

    owner = relationship("User", back_populates="assets")

    def __str__(self):
        attr_list = ", ".join([
            f"owner_id={self.owner_id}",
            f"symbol={self.asset_symbol}",
            f"amount={self.asset_amount}",
        ])
        return f"<User {attr_list}>"

    def __repr__(self):
        return self.__str__()


class Company(Base):
    """Abstracting the individual securities/companies.
    Initial value is the amount of cash put into the company by the founder.
    Market price is the price of a share of this company, which is determined 
    either at founding, or by the latest transaction on this company's stock
    """
    __tablename__ = "companies"

    symbol = Column(String(10), nullable=False, primary_key=True)
    name = Column(String(50), nullable=False)
    initial_value = Column(Float, nullable=False)
    initial_size = Column(Integer, nullable=False)
    founder_id = Column(Integer, ForeignKey('users.user_id', ondelete="SET NULL"))
    market_price = Column(Float, nullable=False)
    create_dttm = Column(DateTime, default=dt.datetime.utcnow)

    founder = relationship("User", back_populates="companies")

    def __str__(self):
        attr_list = ", ".join([
            f"symbol={self.symbol}",
            f"name={self.name}",
            f"founder_id={self.founder_id}",
            f"create_dttm={self.create_dttm}",
        ])
        return f"<User {attr_list}>"

    def __repr__(self):
        return self.__str__()


class Order(Base):
    __tablename__ = 'orders'

    order_id = Column(Integer, primary_key=True)
    security_symbol = Column(String(10), nullable=False)
    side = Column(String(3), nullable=False)
    size = Column(Integer, nullable=False)
    # market orders and sub-orders of market orders do not have target price
    price = Column(Float)
    all_or_none = Column(Boolean, nullable=False, default=False)
    immediate_or_cancel = Column(Boolean, nullable=False, default=False)
    active = Column(Boolean, nullable=False, default=False)
    # good_util_cancelled = Column(Boolean, nullable=False) is out of scope
    parent_order_id = Column(Integer, unique=True)
    owner_id = Column(Integer, ForeignKey('users.user_id', ondelete="CASCADE"))
    cancelled_dttm = Column(DateTime)
    create_dttm = Column(DateTime, default=dt.datetime.utcnow)

    owner = relationship("User", back_populates="orders")

    # A numerical placeholder, not a part of the database schema
    remaining_size: int = 0

    def create_suborder(self):
        return Order(
            security_symbol=self.security_symbol,
            side=self.side,
            size=self.remaining_size,
            price=self.price,
            all_or_none=self.all_or_none,
            immediate_or_cancel=self.immediate_or_cancel,
            active=self.active,
            parent_order_id=self.order_id,
            owner_id=self.owner_id,
            cancelled_dttm=self.cancelled_dttm,
            create_dttm=self.create_dttm
        )

    def copy(self):
        """Return an Order object with exactly the same set of attributes; 
        useful when synchronizing entries across two different ORM sessions

        :return: [description]
        :rtype: [type]
        """
        return Order(
            order_id=self.order_id,
            security_symbol=self.security_symbol,
            side=self.side,
            size=self.size,
            price=self.price,
            all_or_none=self.all_or_none,
            immediate_or_cancel=self.immediate_or_cancel,
            active=self.active,
            parent_order_id=self.parent_order_id,
            owner_id=self.owner_id,
            cancelled_dttm=self.cancelled_dttm,
            create_dttm=self.create_dttm
        )

    @property
    def json(self) -> str:
        """Return a JSON string that captures a manually specified set of 
        instance attributes

        :return: the JSON string
        :rtype: str
        """
        return json.dumps({
            'order_id': self.order_id,
            'security_symbol': self.security_symbol,
            'side': self.side,
            'size': self.size,
            'price': self.price,
            'all_or_none': self.all_or_none,
            'immediate_or_cancel': self.immediate_or_cancel,
            'active': self.active,
            'owner_id': self.owner_id,
            'parent_order_id': self.parent_order_id,
            'cancelled_dttm': self.cancelled_dttm
        })
    
    @classmethod
    def from_json(cls, jstring: str):
        """Given a JSON string, return an Order object. Given that this method 
        is defined under the Base model class, there is no way to bind the 
        object to any session, as the ID columns are just integers.
        It will be up to the external session management to recover the exact 
        mapping from the object to the database entry.

        :param jstring: a JSON string
        :type jstring: str
        :return: The Order object that is mapped from the object
        :rtype: Order
        """
        return cls(**json.loads(jstring))

    def __repr__(self):
        attr_list = ", ".join([
            f"id={self.order_id}",
            f"symbol={self.security_symbol}",
            f"side={self.side}",
            f"size={self.size}",
            f"price={self.price}",
            f"owner_id={self.owner_id}",
        ])
        return f"<Order({attr_list})>"

    def __str__(self):
        return self.__repr__()


class Transaction(Base):
    __tablename__ = 'transactions'

    transaction_id = Column(Integer, primary_key=True)
    security_symbol = Column(String(10), nullable=False)
    size = Column(Integer, nullable=False)
    price = Column(Float, nullable=False)
    # With asks and bids, I am not going to define any relationships because 
    # the concept is strange: what is order.transactions? Is it all transactions 
    # that this order is involved in? Or is it specific to when this order is 
    # a ask vs a bid?
    ask_id = Column(Integer, 
        ForeignKey('orders.order_id', ondelete="CASCADE"), nullable=False)
    bid_id = Column(Integer, 
        ForeignKey('orders.order_id', ondelete="CASCADE"), nullable=False)
    aggressor_order_id = Column(Integer, 
        ForeignKey('orders.order_id', ondelete="CASCADE"), nullable=False)
    resting_order_id = Column(Integer,
        ForeignKey('orders.order_id', ondelete="CASCADE"), 
        nullable=False, unique=True)
    transact_dttm = Column(DateTime, nullable=False, default=dt.datetime.utcnow)

    def __repr__(self):
        attr_list = ", ".join([
            f"id={self.transaction_id}",
            f"time={self.transact_dttm}",
            f"price={self.price}",
            f"size={self.size}",
            f"ask_id={self.ask_id}",
            f"bid_id={self.bid_id}",
        ])
        return f"<Transaction({attr_list})>"

    def __str__(self):
        return self.__repr__()


class MatchingEngineLog(Base):
    """A database-side logging table for recording matching engine's activities:

    :param Base: [description]
    :type Base: [type]
    """
    __tablename__ = 'me_logs'

    log_id = Column(Integer, primary_key=True)
    hostname = Column(String(256), nullable=False)
    pid = Column(Integer, nullable=False)
    log_dttm = Column(DateTime, default=dt.datetime.utcnow, nullable=False)
    log_msg = Column(String(1024))
    ext_ref = Column(String(32))
    ext_ref_id = Column(Integer)



import os
from chives.configs.default_config import DEFAULT_CONFIG

def environment_overwrite(config: dict) -> dict:
    """Given a dictionary, return a copy where for each of its key, the value 
    is overwritten by environment variable if there is one

    :param config: the original configuration dictionary
    :type config: dict
    :return: a copy of the original configuration dictionary but overwritten by 
    the 
    :rtype: dict
    """
    return {k: os.getenv(k, v) for k, v in config.items()}



import pathlib
from setuptools import setup, find_packages

HERE = pathlib.Path(__file__).parent 
README = (HERE / "README.md").read_text()

setup(
    name="chives-exchange",
    version="0.2.3",
    description="Stock exchange with a matching engine and webserver",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/xuganyu96/chives-exchange",
    author='Ganyu "Bruce" Xu',
    author_email="xuganyu96@gmail.com",
    packages=find_packages(include=('chives', 'chives.*')),
    install_requires=[
        "Babel==2.9.0",
        "cryptography==3.2.1",
        "Flask==1.1.2",
        "flask-login==0.5.0",
        "flask-wtf==0.14.3",
        "pandas==1.1.4",
        "pika==1.1.0",
        "pymysql==0.10.1",
        "pytest==6.1.2",
        "requests==2.24.0",
        "SQLAlchemy==1.3.20"
        # uwsgi should be used for deployment, but is not an actual requirement
        # of the library, and is thus not included.
    ],
    entry_points={
        'console_scripts': ['chives=chives.__main__:main']
    },
    include_package_data=True
)



from collections import namedtuple
import datetime as dt
from math import floor
import random
import typing as ty

from flask import Blueprint, jsonify, request
from flask_login import login_required, current_user
import pandas as pd

from chives.db import get_db
from chives.models import Company, Transaction

CandleStickDataPoint = namedtuple(
    # Respectively: dttm, open, high, low, close
    "CandleStickDataPoint", ["t", "o", "h", "l", "c"])
ZoomConfig = namedtuple("ZoomConfig", ['cutoff_offset', 'scale_unit', 'agg_tspan'])
UNIX_START = dt.datetime(1970, 1, 1, 0, 0, 0)
ZOOM_CONFIGS = {
    'day': ZoomConfig(dt.timedelta(hours=24), "hour", dt.timedelta(minutes=10)),
    'month': ZoomConfig(dt.timedelta(days=30), "day", dt.timedelta(hours=6)),
    'year': ZoomConfig(dt.timedelta(days=365), "month", dt.timedelta(days=1)),
    'max': ZoomConfig(dt.timedelta(days=3650), "year", dt.timedelta(days=7))
}

bp = Blueprint("api", __name__, url_prefix="/api")

@bp.route("/is_authenticated", methods=("GET",))
@login_required
def is_authenticated():
    return jsonify(True)

@bp.route("/autocomplete_companies", methods=("GET",))
@login_required 
def autocomplete_companies():
    db = get_db()
    companies = db.query(Company).all()
    data = {c.symbol: None for c in companies}
    return jsonify(data)


@bp.route("/stock_chart_data", methods=("GET",))
@login_required 
def stock_chart_data():
    symbol = request.args['symbol']
    zoom = request.args['zoom'] if "zoom" in request.args else "year"
    debug = int(request.args['debug']) if "debug" in request.args else 0
    db = get_db()

    # Query all transactions with transaction dttm sorted from earlier to later
    tfilter = (Transaction.security_symbol == symbol)
    # Add the dttm filter, which depends on the zoom level
    cutoff = dt.datetime.utcnow() - ZOOM_CONFIGS[zoom].cutoff_offset
    scale_unit = ZOOM_CONFIGS[zoom].scale_unit
    tfilter = tfilter & (Transaction.transact_dttm >= cutoff)

    sort_key = Transaction.transact_dttm.asc()
    query = db.query(Transaction).filter(tfilter).order_by(sort_key)

    # Convert the set of transactions into a 2 column dataframe: price vs dttm
    df = pd.read_sql(query.statement, db.bind)[['transact_dttm', 'price']]
    if debug:
        # If debug is set to True, then return dummy data without reading
        # from database
        random.seed(10)
        df = pd.DataFrame({
            "transact_dttm": [
                dt.datetime.fromtimestamp(
                    random.uniform(
                        cutoff.timestamp(),
                        dt.datetime.utcnow().timestamp()
                    )
                )  for i in range(500)],
            "price": [random.uniform(10, 100) for i in range(500)]
        }).sort_values('transact_dttm')
    agg_data_points = aggregate_stock_chart(df, zoom)
    agg_data_points_dict = [{
        "t": dp.t * 1000,
        "o": dp.o,
        "h": dp.h,
        "l": dp.l,
        "c": dp.c,
        "y": dp.c
    } for dp in agg_data_points]
    max_price = 0 if pd.isna(df['price'].max()) else df['price'].max()
    min_price = 0 if pd.isna(df['price'].min()) else df['price'].min()
    price_std = 0 if pd.isna(df['price'].std()) else df['price'].std()
    
    data = {
        "labels": [],
        "datasets": [{
            "label": "market price",
            "data": agg_data_points_dict
        }]
    }
    options = {
        "scales": {
            "x": {
                "type": 'time',
                "distribution": 'linear',
                "time": {"unit": scale_unit}
            },
            "y": {
                "suggestedMax": max_price + 0.7 * price_std,
                "suggestedMin": max(0, min_price - 0.7 * price_std)
            }
        }
    }
    return jsonify({"data": data, "options": options})


def aggregate_stock_chart(df: pd.DataFrame, 
    zoom: str = "day") -> ty.List[ty.Dict]:
    """Given a DataFrame that contains two non-null columns "transact_dttm" and 
    "price", return a list of CandleStickDataPoint instances that represent the 
    aggregated trade data of the time interval that starts with dp.T and ends 
    with the next dp.T

    :param df: [description]
    :type df: pd.DataFrame
    :param zoom: [description], defaults to "day"
    :type zoom: str, optional
    """
    if df.shape[0] == 0:
        # If there is no transaction, then return an empty list
        return []
    
    assert zoom in ["day", "month", "year", "max"]
    agg_tspan = ZOOM_CONFIGS[zoom].agg_tspan
    
    global_min_transact_ts = df['transact_dttm'].min()
    global_agg_start = (global_min_transact_ts - UNIX_START) // agg_tspan \
        * agg_tspan + UNIX_START
    df['bucket_idx'] = (df['transact_dttm'] - global_agg_start) / agg_tspan
    df['bucket_idx'] = df['bucket_idx'].apply(floor)

    output = []
    # an interval's close should be next interval's open
    prior_close = None
    for idx in df['bucket_idx'].unique():
        part = df.loc[df['bucket_idx'] == idx]
        # If there is a prior close, then open is set to prior close
        open = prior_close if prior_close else round(
            part.sort_values('transact_dttm').head(1)['price'].values[0], 2)
        # Update prior close
        prior_close = close = round(
            part.sort_values('transact_dttm').tail(1)['price'].values[0], 2)
        high = round(part['price'].max(), 2)
        low = round(part['price'].min(), 2)
        ts = (global_agg_start + idx * agg_tspan).timestamp()
        output.append(CandleStickDataPoint(
            t=ts, o=open, h=high, l=low, c=close
        ))
    
    return output
    


from sqlalchemy import create_engine

from chives.cli import parser as chives_parser
from chives.matchingengine import start_engine
from chives.models import Base
from chives.webserver import create_app


def main():
    args = chives_parser.parse_args()

    if args.subcommand == "start_engine":
        sql_engine = create_engine(args.sql_uri, echo=args.verbose)
        start_engine({
            "SQLALCHEMY_CONN": args.sql_uri,
            "SQLALCHEMY_ECHO": args.verbose
        })
    if args.subcommand == "initdb":
        sql_engine = create_engine(args.sql_uri, echo=args.verbose)
        Base.metadata.create_all(sql_engine)
    if args.subcommand == "webserver":
        # Obtain the agruments that form the application configuration, then 
        # pass the configuration into the app factory before running the app
        sql_uri = args.sql_uri
        config={
            "SQLALCHEMY_CONN": args.sql_uri,
            "SQLALCHEMY_ECHO": args.verbose
        }
        
        app = create_app(config)
        app.run(port=args.webserver_port, debug=args.debug)


if __name__ == "__main__":
    main()
